# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
import datetime
import logging
import os
from pathlib import Path
import shutil
import signal
import sys
import traceback

from colcon_core.environment_variable import EnvironmentVariable

# a custom environment variable is necessary since PYTHONWARNINGS doesn't
# support passing a regular expression for the module entry
# see https://bugs.python.org/issue34624
"""Environment variable to set the warnings filter for colcon modules"""
WARNINGS_ENVIRONMENT_VARIABLE = EnvironmentVariable(
    'COLCON_WARNINGS',
    'Set the warnings filter similar to PYTHONWARNINGS except that the module '
    "entry is implicitly set to 'colcon.*'")

warnings_filters = os.environ.get(WARNINGS_ENVIRONMENT_VARIABLE.name)
if warnings_filters:
    import warnings
    # filters are separated by commas
    for f in warnings_filters.split(','):
        # fields are separated by colons
        fields = f.split(':', 4)
        if len(fields) < 5:
            fields += [''] * (5 - len(fields))
        action, message, category, module, line = fields
        try:
            category = warnings._getcategory(category)
        except Exception:
            print(
                "The category field '{category}' must be a valid warnings "
                'class name'.format_map(locals()), file=sys.stderr)
            sys.exit(1)
        if module:
            print(
                'The module field of the {WARNINGS_ENVIRONMENT_VARIABLE.name} '
                'filter should be empty, otherwise use PYTHONWARNINGS instead'
                .format_map(locals()), file=sys.stderr)
            sys.exit(1)
        warnings.filterwarnings(
            action, message=message, category=category or Warning,
            module='colcon.*', lineno=line if line else 0)

from colcon_core.argument_parser import decorate_argument_parser  # noqa: E402 E501 I100 I202
from colcon_core.argument_parser import SuppressUsageOutput  # noqa: E402
from colcon_core.entry_point import load_entry_points  # noqa: E402
from colcon_core.location import create_log_path  # noqa: E402
from colcon_core.location import get_log_path  # noqa: E402
from colcon_core.location import set_default_config_path  # noqa: E402
from colcon_core.location import set_default_log_path  # noqa: E402
from colcon_core.logging import add_file_handler  # noqa: E402
from colcon_core.logging import colcon_logger  # noqa: E402
from colcon_core.logging import get_numeric_log_level  # noqa: E402
from colcon_core.logging import set_logger_level_from_env  # noqa: E402
from colcon_core.plugin_system import get_first_line_doc  # noqa: E402
from colcon_core.verb import get_verb_extensions  # noqa: E402

"""Environment variable to set the log level"""
LOG_LEVEL_ENVIRONMENT_VARIABLE = EnvironmentVariable(
    'COLCON_LOG_LEVEL',
    'Set the log level (debug|10, info|20, warn|30, error|40, critical|50, or '
    'any other positive numeric value)')

"""Environment variable to set the configuration directory"""
HOME_ENVIRONMENT_VARIABLE = EnvironmentVariable(
    'COLCON_HOME',
    'Set the configuration directory (default: ~/.colcon)')


_command_exit_handlers = []


def register_command_exit_handler(handler):
    """
    Register a callable to be invoked after the command finished.

    Repeated registrations of the same handler are ignored.

    :param handler: The callable
    """
    global _command_exit_handlers
    if handler not in _command_exit_handlers:
        _command_exit_handlers.append(handler)


def main(*, command_name='colcon', argv=None):
    """
    Execute the main logic of the command.

    The overview of the process:
    * Configure logging level based on an environment variable
    * Configure the configuration path
    * Create the argument parser

      * Document all environment variables
      * Decorate the parsers with additional functionality
      * Add the available verbs and their arguments

    * Configure logging level based on an arguments
    * Create an invocation specific log directory
    * Invoke the logic of the selected verb (if applicable)
    * Invoke registered exit handlers in reverse order

    :param str command_name: The name of the command invoked
    :param list argv: The list of arguments
    :returns: The return code
    """
    global _command_exit_handlers
    try:
        return _main(command_name=command_name, argv=argv)
    except KeyboardInterrupt:
        return signal.SIGINT
    finally:
        # invoke all exit handlers
        while _command_exit_handlers:
            handler = _command_exit_handlers.pop()
            handler()


def _main(*, command_name, argv):
    global colcon_logger
    # default log level, for searchability: COLCON_LOG_LEVEL
    colcon_logger.setLevel(logging.WARNING)
    set_logger_level_from_env(
        colcon_logger, '{command_name}_LOG_LEVEL'.format_map(locals()).upper())
    colcon_logger.debug(
        'Command line arguments: {argv}'
        .format(argv=argv if argv is not None else sys.argv))

    # set default locations for config files, for searchability: COLCON_HOME
    set_default_config_path(
        path=(
            Path('~') / '.{command_name}'.format_map(locals())).expanduser(),
        env_var='{command_name}_HOME'.format_map(locals()).upper())

    parser = create_parser('colcon_core.environment_variable')

    verb_extensions = get_verb_extensions()

    # add subparsers for all verb extensions but without arguments for now
    subparser = create_subparser(
        parser, command_name, verb_extensions, attribute='verb_name')
    verb_parsers = add_parsers_without_arguments(
        parser, subparser, verb_extensions, attribute='verb_name')

    with SuppressUsageOutput([parser] + list(verb_parsers.values())):
        known_args, _ = parser.parse_known_args(args=argv)

    # add the arguments for the requested verb
    if known_args.verb_name:
        add_parser_arguments(known_args.verb_parser, known_args.verb_extension)

    args = parser.parse_args(args=argv)
    context = CommandContext(command_name=command_name, args=args)

    if args.log_level:
        # the value might be provided externally and needs to be checked again
        colcon_logger.setLevel(get_numeric_log_level(args.log_level))

    colcon_logger.debug(
        'Parsed command line arguments: {args}'.format_map(locals()))

    # error: no verb provided
    if args.verb_name is None:
        print(parser.format_usage())
        return 'Error: No verb provided'

    # set default locations for log files, for searchability: COLCON_LOG_PATH
    now = datetime.datetime.now()
    now_str = str(now)[:-7].replace(' ', '_').replace(':', '-')
    set_default_log_path(
        base_path=args.log_base,
        env_var='{command_name}_LOG_PATH'.format_map(locals()).upper(),
        subdirectory='{args.verb_name}_{now_str}'.format_map(locals()))

    # add a file handler writing all levels if logging isn't disabled
    log_path = get_log_path()
    if log_path is not None:
        create_log_path(args.verb_name)
        handler = add_file_handler(
            colcon_logger, log_path / 'logger_all.log')
        # write previous log messages to the file handler
        log_record = colcon_logger.makeRecord(
            colcon_logger.name, logging.DEBUG, __file__, 0,
            'Command line arguments: {argv}'
            .format(argv=argv if argv is not None else sys.argv),
            None, None)
        handler.handle(log_record)
        log_record = colcon_logger.makeRecord(
            colcon_logger.name, logging.DEBUG, __file__, 0,
            'Parsed command line arguments: {args}'.format_map(locals()),
            None, None)
        handler.handle(log_record)

    # set an environment variable named after the command (if not already set)
    # which allows subprocesses to identify they are invoked by this command
    if command_name.upper() not in os.environ:
        os.environ[command_name.upper()] = '1'

    # invoke verb
    return verb_main(context, colcon_logger)


def create_parser(environment_variables_group_name):
    """
    Create the argument parser.

    It uses a custom raw description help formatter to maintain newlines.
    It uses the available argument parser extensions to decorate the parsers.
    It enumerates the registered environment variables in the epilog of the
    help message.

    :param str environment_variables_group_name: The entry point group name for
      the environment variable extensions
    :returns: The argument parser
    """
    # workaround a limitation in argparse to accept arguments to options
    # which begin with a dash but are not options themselves
    # https://bugs.python.org/issue9334
    class CustomArgumentParser(argparse.ArgumentParser):

        def _parse_optional(self, arg_string):
            result = super()._parse_optional(arg_string)
            if result == (None, arg_string, None):
                # in the case there the arg is classified as an unknown 'O'
                # override that and classify it as an 'A'
                return None
            return result

    # top level parser
    parser = CustomArgumentParser(
        prog=get_prog_name(),
        formatter_class=CustomFormatter,
        epilog=(
            get_environment_variables_epilog(
                environment_variables_group_name
            ) + '\n\n' + READTHEDOCS_MESSAGE))

    # enable introspecting and intercepting all command line arguments
    parser = decorate_argument_parser(parser)

    add_log_level_argument(parser)

    return parser


def get_prog_name():
    """Get the prog name used for the argparse parser."""
    prog = sys.argv[0]
    basename = os.path.basename(prog)
    if basename == '__main__.py':
        # use the module name in case the script was invoked with python -m ...
        prog = os.path.basename(os.path.dirname(prog))
    elif shutil.which(basename) == prog:
        # use basename only if it is on the PATH
        prog = basename
    return prog


class CustomFormatter(argparse.RawDescriptionHelpFormatter):
    """A custom formatter to maintain newlines."""

    def _split_lines(self, text, width):
        """Maintain newlines when the text starts with 'r|'."""
        lines = []
        for line in text.splitlines():
            if len(line) <= width:
                lines.append(line)
            else:
                lines += super()._split_lines(line, width)
        return lines


def get_environment_variables_epilog(group_name):
    """
    Get a message enumerating the registered environment variables.

    :param str group_name: The entry point group name for the environment
      variable extensions
    :returns: The message for the argument parser epilog
    :rtype: str
    """
    # list environment variables with descriptions
    entry_points = load_entry_points(group_name)
    env_vars = {
        env_var.name: env_var.description for env_var in entry_points.values()}
    epilog_lines = []
    for name in sorted(env_vars.keys()):
        epilog_lines += _format_pair(name, env_vars[name], indent=2, align=24)
    return 'Environment variables:\n' + '\n'.join(epilog_lines)


READTHEDOCS_MESSAGE = 'For more help and usage tips, see ' \
    'https://colcon.readthedocs.io'


def add_log_level_argument(parser):
    """
    Add the argument for the log level to the parser.

    :param parser: The argument parser
    """
    parser.add_argument(
        '--log-base',
        help='The base path for all log directories (default: ./log, to '
             'disable: {os.devnull})'.format_map(globals()))
    parser.add_argument(
        '--log-level', action=LogLevelAction,
        help='Set log level for the console output, either by numeric or '
             'string value (default: warning)')


class LogLevelAction(argparse.Action):
    """Accept either positive integers or known log level names."""

    def __init__(self, option_strings, dest, *, nargs=None, **kwargs):
        """See :class:`argparse.Action.__init__`."""
        if nargs is not None:  # pragma: no cover
            raise ValueError('nargs not allowed')
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        """See :class:`argparse.Action.__call__`."""
        try:
            value = get_numeric_log_level(values)
        except ValueError as e:  # noqa: F841
            parser.error(
                '{option_string} has unsupported value, {e}'
                .format_map(locals()))
        setattr(namespace, self.dest, value)


def add_subparsers(parser, cmd_name, verb_extensions, *, attribute):
    """
    Create argparse subparsers for each verb.

    The `cmd_name` is used for the title and description of the argparse
    `add_subparsers` function call.

    For each verb extension a subparser is created.
    If the extension has an `add_arguments` method it is being called with the
    subparser being passed as the only argument.

    :param parser: The argument parser for this command
    :param str cmd_name: The name of the command to which the verbs are being
      added
    :param dict verb_extensions: The verb extensions indexed by the verb name
    :param str attribute: The name of the attribute in the parsed args for the
      selected verb
    """
    subparser = create_subparser(
        parser, cmd_name, verb_extensions, attribute=attribute)
    verb_parsers = add_parsers_without_arguments(
        parser, subparser, verb_extensions, attribute=attribute)
    for name, verb_parser in verb_parsers.items():
        add_parser_arguments(
            verb_parser, verb_extensions[name])


def create_subparser(parser, cmd_name, verb_extensions, *, attribute):
    """
    Create the special action object to contain subparsers.

    The `cmd_name` is used for the title and description of the argparse
    `add_subparsers` function call.

    :param parser: The argument parser for this command
    :param str cmd_name: The name of the command to which the verbs are being
      added
    :param dict verb_extensions: The verb extensions indexed by the verb name
    :param str attribute: The name of the attribute in the parsed args for the
      selected verb
    :returns: The special action object
    """
    global colcon_logger
    assert verb_extensions, 'No verb extensions'

    # list of available verbs with their descriptions
    verbs = []
    for name, extension in verb_extensions.items():
        verbs += _format_pair(
            name, get_first_line_doc(extension), indent=0, align=22)

    # add subparser with description of verb extensions
    subparser = parser.add_subparsers(
        title='{cmd_name} verbs'.format_map(locals()),
        description='\n'.join(verbs),
        dest=attribute,
        help='call `{cmd_name} VERB -h` for specific help'
             .format_map(locals())
    )
    return subparser


def add_parsers_without_arguments(
    parser, subparser, verb_extensions, *, attribute
):
    """
    Create subparsers for each verb but without any arguments.

    For each verb extension a subparser is created.

    :param parser: The argument parser for this command
    :param subparser: The special action object to add the subparsers to
    :param dict verb_extensions: The verb extensions indexed by the verb name
    :param str attribute: The name of the attribute in the extension containing
      the verb
    :returns: The subparsers indexed by the verb name
    :rtype: dict
    """
    verb_parsers = {}
    # add verb specific group and arguments
    for name, extension in verb_extensions.items():
        verb_parser = subparser.add_parser(
            getattr(extension, attribute.upper()),
            description=get_first_line_doc(extension) + '.',
            formatter_class=parser.formatter_class,
        )
        verb_parser.set_defaults(
            verb_parser=verb_parser, verb_extension=extension,
            main=extension.main)
        verb_parsers[name] = verb_parser
    return verb_parsers


def add_parser_arguments(verb_parser, extension):
    """
    Add the arguments and recursive subparsers to a specific verb parser.

    If the extension has an `add_arguments` method it is being called with the
    subparser being passed as the only argument.

    :param verb_parser: The verb parser
    :param extension: The verb extension
    """
    if hasattr(extension, 'add_arguments'):
        retval = extension.add_arguments(parser=verb_parser)
        if retval is not None:
            colcon_logger.error(
                "Exception in verb extension '{extension.VERB_NAME}': "
                'add_arguments() should return None'.format_map(locals()))


def _format_pair(key, value, *, indent, align):
    """
    Format a key value pair to align with others printed by argparse.

    :param str key: The key
    :param str value: The value
    :param int indent: The indentation level of the key
    :param int align: The indentation level of the value
    :returns: The indented and potentially wrapped line(s)
    :rtype: str
    """
    lines = []
    prefix = ' ' * indent + key
    # wrap between key and value if the gap between them is smaller than this
    minimum_gap = 2
    if len(prefix) + minimum_gap <= align:
        # key fits in the same line as the value
        lines.append(prefix + ' ' * (align - len(prefix)))
    else:
        # the key is too long, the value needs to start on the next line
        lines.append(prefix)
        lines.append(' ' * align)

    # wrap between key and value if the gap between them is smaller than this
    maximum_line_length = 80
    maximum_value_length = maximum_line_length - align
    while value:
        if len(value) > maximum_value_length:
            try:
                # look for a space within the desired length
                i = value.rindex(' ', 0, maximum_value_length)
            except ValueError:
                # no space to wrap, just append everything in a single line
                pass
            else:
                # append part to last line
                lines[-1] += value[0:i]
                value = value[i + 1:].lstrip()
                # start a new line with the spaces for the alignment
                lines.append(' ' * align)
                continue

        # either the remaining value is short enough or no space was found
        lines[-1] += value
        break
    return lines


class CommandContext:
    """The context providing the command name and the parsed arguments."""

    __slots__ = ('command_name', 'args')

    def __init__(self, *, command_name: str, args: object):  # noqa: D107
        self.command_name = command_name
        self.args = args


def verb_main(context, logger):
    """
    Invoke the logic of the selected verb.

    If the invocation is interrupted the returned error code is
    `signal.SIGINT`.
    If the verb raises a `RuntimeException` an error message is logged which
    contains the message of the exception.
    For any other exception a traceback is included in the logged error
    message.

    :param context: The :class:`CommandContext`
    :param logger: The logger
    :returns: The return code
    """
    # call the extension's main method
    try:
        # catch exceptions raised in verb extension
        rc = context.args.main(context=context)
    except RuntimeError as e:  # noqa: F841
        # only log the error message for "known" exceptions
        logger.error(
            '{context.command_name} {context.args.verb_name}: {e}'
            .format_map(locals()))
        return 1
    except Exception as e:  # noqa: F841
        # log the error message and a traceback for "unexpected" exceptions
        exc = traceback.format_exc()
        logger.error(
            '{context.command_name} {context.args.verb_name}: {e}\n{exc}'
            .format_map(locals()))
        return 1
    return rc



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import sys

from colcon_core.event.output import StderrLine
from colcon_core.event.output import StdoutLine
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version


class ConsoleDirectEventHandler(EventHandlerExtensionPoint):
    """
    Pass output directly to stdout/err.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.output.StdoutLine`
    - :py:class:`colcon_core.event.output.StderrLine`
    """

    # this handler is enabled by default
    # but other handlers might choose to change that presetting
    ENABLED_BY_DEFAULT = True

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self.enabled = ConsoleDirectEventHandler.ENABLED_BY_DEFAULT

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, StdoutLine):
            if isinstance(data.line, bytes):
                sys.stdout.buffer.write(data.line)
            else:
                sys.stdout.write(data.line)
            sys.stdout.flush()

        elif isinstance(data, StderrLine):
            if isinstance(data.line, bytes):
                sys.stderr.buffer.write(data.line)
            else:
                sys.stderr.write(data.line)
            sys.stderr.flush()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from io import StringIO
import os

from colcon_core.logging import colcon_logger
try:
    from em import Interpreter
    from em import OVERRIDE_OPT
except ImportError as e:
    try:
        import em  # noqa: F401
    except ImportError:
        e.msg += " The Python package 'empy' must be installed"
        raise e from None
    e.msg += " The Python package 'empy' must be installed and 'em' must " \
        'not be installed since both packages share the same namespace'
    raise e from None

logger = colcon_logger.getChild(__name__)


def expand_template(template_path, destination_path, data):
    """
    Expand an EmPy template.

    The directory of the destination path is created if necessary.

    :param template_path: The patch of the template file
    :param destination_path: The path of the generated expanded file
    :param dict data: The data used for expanding the template
    :raises: Any exception which `em.Interpreter.string` might raise
    """
    output = StringIO()
    try:
        # disable OVERRIDE_OPT to avoid saving / restoring stdout
        interpreter = CachingInterpreter(
            output=output, options={OVERRIDE_OPT: False})
        with template_path.open('r') as h:
            content = h.read()
        interpreter.string(content, str(template_path), locals=data)
        output = output.getvalue()
    except Exception as e:  # noqa: F841
        logger.error(
            "{e.__class__.__name__} processing template '{template_path}'"
            .format_map(locals()))
        raise
    else:
        os.makedirs(str(destination_path.parent), exist_ok=True)
        # if the destination_path is a symlink remove the symlink
        # to avoid writing to the symlink destination
        if destination_path.is_symlink():
            destination_path.unlink()
        with destination_path.open('w') as h:
            h.write(output)
    finally:
        interpreter.shutdown()


class BypassStdoutInterpreter(Interpreter):
    """Interpreter for EmPy which keeps `stdout` unchanged."""

    def installProxy(self):  # noqa: D102 N802
        # avoid replacing stdout with ProxyFile
        pass


cached_tokens = {}


class CachingInterpreter(BypassStdoutInterpreter):
    """Interpreter for EmPy which which caches parsed tokens."""

    def parse(self, scanner, locals=None):  # noqa: A002 D102
        global cached_tokens
        data = scanner.buffer
        # try to use cached tokens
        tokens = cached_tokens.get(data)
        if tokens is None:
            # collect tokens and cache them
            tokens = []
            while True:
                token = scanner.one()
                if token is None:
                    break
                tokens.append(token)
            cached_tokens[data] = tokens

        # reimplement the parse method using the (cached) tokens
        self.invoke('atParse', scanner=scanner, locals=locals)
        for token in tokens:
            self.invoke('atToken', token=token)
            token.run(self, locals)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import defaultdict
import traceback

from colcon_core.logging import colcon_logger
from colcon_core.package_augmentation import augment_packages
from colcon_core.package_discovery import add_package_discovery_arguments
from colcon_core.package_discovery import discover_packages
from colcon_core.package_identification \
    import get_package_identification_extensions
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name
from colcon_core.topological_order import topological_order_packages

logger = colcon_logger.getChild(__name__)


class PackageSelectionExtensionPoint:
    """
    The interface for package selection extensions.

    A package selection extension determines the subset of packages to be
    processed.

    For each instance the attribute `PACKAGE_SELECTION_NAME` is being set to
    the basename of the entry point registering the extension.
    """

    """The version of the package selection extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the package selection.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def check_parameters(self, *, args, pkg_names):
        """
        Check is the passed arguments have valid values.

        The method is intended to be overridden in a subclass.
        It should either warn about invalid values and gracefully continue or
        raise a `SystemExit` exception.

        :param args: The parsed command line arguments
        :param pkg_names: The set of package names
        """
        pass

    def select_packages(self, *, args, decorators):
        """
        Identify the packages which should be skipped.

        By default all package decorators are marked as "selected".

        This method must be overridden in a subclass.

        :param args: The parsed command line arguments
        :param list decorators: The package decorators in topological order
        """
        raise NotImplementedError()


def add_arguments(parser):
    """
    Add the command line arguments for the package selection extensions.

    The function will call :function:`add_package_discovery_arguments` to add
    the package discovery arguments.

    :param parser: The argument parser
    """
    add_package_discovery_arguments(parser)

    _add_package_selection_arguments(parser)


def get_package_selection_extensions():
    """
    Get the available package selection extensions.

    The extensions are ordered by their entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_SELECTION_NAME = name
    return order_extensions_by_name(extensions)


def _add_package_selection_arguments(parser):
    """
    Add the command line arguments for the package selection extensions.

    :param parser: The argument parser
    """
    package_selection_extensions = get_package_selection_extensions()
    group = parser.add_argument_group(title='Package selection arguments')
    for extension in package_selection_extensions.values():
        try:
            retval = extension.add_arguments(parser=group)
            assert retval is None, 'add_arguments() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in package selection extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package selection extension '
                "'{extension.PACKAGE_SELECTION_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one


def get_packages(
    args, *,
    additional_argument_names=None,
    direct_categories=None, recursive_categories=None
):
    """
    Get the selected package decorators in topological order.

    The overview of the process:
    * Get the package descriptors
    * Order them topologically
    * Select the packages based on the command line arguments

    :param additional_argument_names: A list of additional arguments to
      consider
    :param Iterable[str] direct_categories: The names of the direct categories
    :param Iterable[str] recursive_categories: The names of the recursive
      categories
    :rtype: list
    :raises RuntimeError: if the returned set of packages contains duplicates
      package names
    """
    descriptors = get_package_descriptors(
        args, additional_argument_names=additional_argument_names)
    decorators = topological_order_packages(
        descriptors,
        direct_categories=direct_categories,
        recursive_categories=recursive_categories)
    select_package_decorators(args, decorators)

    # check for duplicate package names
    pkgs = [m.descriptor for m in decorators if m.selected]
    if len({d.name for d in pkgs}) < len(pkgs):
        pkg_paths = defaultdict(list)
        for d in pkgs:
            pkg_paths[d.name].append('  - {d.path}'.format_map(locals()))
        raise RuntimeError(
            'Duplicate package names not supported:\n' +
            '\n'.join(
                ('- ' + name + ':\n' + '\n'.join(sorted(pkg_paths[name])))
                for name in sorted(pkg_paths.keys())
                if len(pkg_paths[name]) > 1))

    return decorators


def get_package_descriptors(args, *, additional_argument_names=None):
    """
    Get the package descriptors.

    The overview of the process:
    * Discover the package descriptors using the package discovery and
      identification extensions
    * Check is the passed package selection arguments have valid values
    * Augment the package descriptors

    :param additional_argument_names: A list of additional arguments to
      consider
    :returns: set of
      :py:class:`colcon_core.package_descriptor.PackageDescriptor`
    :rtype: set
    """
    extensions = get_package_identification_extensions()
    descriptors = discover_packages(args, extensions)

    pkg_names = {d.name for d in descriptors}
    _check_package_selection_parameters(args, pkg_names)

    augment_packages(
        descriptors, additional_argument_names=additional_argument_names)
    return descriptors


def _check_package_selection_parameters(args, pkg_names):
    package_selection_extensions = get_package_selection_extensions()
    for extension in package_selection_extensions.values():
        try:
            retval = extension.check_parameters(args=args, pkg_names=pkg_names)
            assert retval is None, 'check_parameters() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in package selection extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package selection extension '
                "'{extension.PACKAGE_SELECTION_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one


def select_package_decorators(args, decorators):
    """
    Select the package decorators based on the command line arguments.

    The `selected` attribute of each decorator is updated by this function.

    :param args: The parsed command line arguments
    :param list decorators: The package decorators in topological order
    """
    # filtering must happen after the topological ordering since otherwise
    # packages in the middle of the dependency graph might be missing
    package_selection_extensions = get_package_selection_extensions()
    for extension in package_selection_extensions.values():
        try:
            retval = extension.select_packages(
                args=args, decorators=decorators)
            assert retval is None, 'select_packages() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in package selection extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package selection extension '
                "'{extension.PACKAGE_SELECTION_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.topological_order import topological_order_packages
import pytest


def test_topological_order_packages():
    d1 = PackageDescriptor('/some/path')
    d1.name = 'a'
    d1.dependencies['build'].add('c')
    d2 = PackageDescriptor('/other/path')
    d2.name = 'b'
    d2.dependencies['run'].add('c')

    d3 = PackageDescriptor('/another/path')
    d3.name = 'c'
    d3.dependencies['build'].add('e')
    d3.dependencies['run'].add('f')
    d3.dependencies['test'].add('d')

    d4 = PackageDescriptor('/yet-another/path')
    d4.name = 'd'
    d4.dependencies['run'].add('f')
    d5 = PackageDescriptor('/more/path')
    d5.name = 'e'
    d5.dependencies['run'].add('f')

    d6 = PackageDescriptor('/yet-more/path')
    d6.name = 'f'

    decos = topological_order_packages(
        {d1, d2, d3, d4, d5, d6})
    names = [d .descriptor.name for d in decos]
    assert names == ['f', 'd', 'e', 'c', 'a', 'b']

    # ensure that input order doesn't affect the result
    decos = topological_order_packages(
        {d6, d5, d4, d3, d2, d1})
    names = [d .descriptor.name for d in decos]
    assert names == ['f', 'd', 'e', 'c', 'a', 'b']


def test_topological_order_packages_with_circular_dependency():
    d1 = PackageDescriptor('/some/path')
    d1.name = 'one'
    d1.dependencies['run'].add('two')

    d2 = PackageDescriptor('/other/path')
    d2.name = 'two'
    d2.dependencies['run'].add('three')

    d3 = PackageDescriptor('/another/path')
    d3.name = 'three'
    d3.dependencies['run'].add('one')
    d3.dependencies['run'].add('six')

    d4 = PackageDescriptor('/yet-another/path')
    d4.name = 'four'

    d5 = PackageDescriptor('/more/path')
    d5.name = 'five'
    d5.dependencies['run'].add('four')

    d6 = PackageDescriptor('/yet-more/path')
    d6.name = 'six'

    with pytest.raises(RuntimeError) as e:
        topological_order_packages({d1, d2, d3, d4})
    lines = str(e.value).splitlines()
    assert len(lines) == 4
    assert lines[0] == 'Unable to order packages topologically:'
    assert lines[1] == "one: ['three', 'two']"
    assert lines[2] == "three: ['one', 'two']"
    assert lines[3] == "two: ['one', 'three']"



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0


class TestFailure:
    """An event containing the identifier of the job with test failures."""

    __slots__ = ('identifier', )

    def __init__(self, identifier):
        """
        Construct a TestFailure.

        :param str identifier: The job identifier
        """
        self.identifier = identifier



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import logging
import os

logging.basicConfig()
colcon_logger = logging.getLogger('colcon')

try:
    import coloredlogs
except ImportError:  # pragma: no cover
    pass
else:
    log_format = os.environ.get(
        'COLOREDLOGS_LOG_FORMAT', '%(name)s %(levelname)s %(message)s')
    coloredlogs.install(level=1, logger=colcon_logger, fmt=log_format)


def set_logger_level_from_env(logger, env_name):
    """
    Set the log level based on an environment variable.

    A warning message is logged if the environment variable has an unsupported
    value.

    :param logger: The logger
    :param str env_var: The name of the environment variable
    """
    log_level = os.environ.get(env_name)
    if log_level:
        try:
            numeric_log_level = get_numeric_log_level(log_level)
        except ValueError as e:  # noqa: F841
            logger.warning(
                "environment variable '{env_name}' has unsupported value "
                "'{log_level}', {e}".format_map(locals()))
        else:
            logger.setLevel(numeric_log_level)


def get_numeric_log_level(value):
    """
    Convert a log level into a numeric value.

    :param value: The log level can be either a string (case insensitive) or a
      positive number
    :returns: The numeric value
    :rtype: int
    :raises ValueError: if the log level string is not one of the valid names
      (`CRITICAL`, `ERROR`, `WARNING`, `INFO`, `DEBUG`) or if the numeric value
      is zero or negative
    """
    try:
        value = int(value)
    except ValueError:
        string_value = value.upper()
        value = logging.getLevelName(string_value)
        if value == 'Level ' + string_value:
            raise ValueError(
                'valid names are: CRITICAL, ERROR, WARNING, INFO, DEBUG '
                '(case-insensitive)')
    else:
        if value < 1:
            raise ValueError('numeric log levels must be positive')
    return value


def add_file_handler(logger, path):
    """
    Add a file handler to the logger which logs messages of all levels.

    :param logger: The logger to add the file handler to
    :param path: The path of the generated log file
    :returns: The added file handler
    :rtype: logging.FileHandler
    """
    class Filter(logging.Filter):

        def __init__(self, ignored_name):
            super().__init__()
            self._ignored_name = ignored_name

        def filter(self, record):  # noqa: A003
            if (
                record.name == self._ignored_name or
                record.name.startswith(self._ignored_name)
            ):
                return 0
            return super().filter(record)

    # get stream handler formatter from root logger to reuse for file handler
    formatter = None
    for handler in logging.getLogger().handlers:
        if isinstance(handler, logging.StreamHandler):
            formatter = handler.formatter
            # filter colcon specific log messages from default stream handler
            handler.addFilter(Filter(colcon_logger.name))

    # add a stream handler replacing the one filtered on the root logger
    handler = logging.StreamHandler()
    if formatter:
        # use same formatter as for stream handler
        handler.setFormatter(formatter)
    handler.setLevel(colcon_logger.getEffectiveLevel())
    colcon_logger.addHandler(handler)

    # add a file handler writing all log levels
    handler = logging.FileHandler(str(path))
    if formatter:
        # if the format string doesn't use the time information
        # prepend the relative time to every message
        if not formatter.usesTime():
            format_message = formatter.formatMessage

            def format_message_with_relative_time(record):
                nonlocal format_message
                return '[%.3fs] ' % (record.created - logging._startTime) + \
                    format_message(record)
            formatter.formatMessage = format_message_with_relative_time
        # use same formatter as for stream handler
        handler.setFormatter(formatter)
    handler.setLevel(1)
    colcon_logger.addHandler(handler)

    # change the logger to handle all levels
    colcon_logger.setLevel(1)

    return handler



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
import os
from pathlib import Path
import sys


def main(argv=sys.argv[1:]):  # noqa: D103
    parser = argparse.ArgumentParser(
        description='Output found packages in topological order')
    parser.add_argument(
        '--merged-install', action='store_true',
        help='All install prefixes are merged into a single location')
    args = parser.parse_args(argv)

    packages = get_packages(Path(__file__).parent, args.merged_install)

    for pkg_name in order_packages(packages):
        print(pkg_name)


def get_packages(prefix_path, merged_install):
    """
    Find packages based on colcon-specific files created during installation.

    :param Path prefix_path: The install prefix path of all packages
    :param bool merged_install: The flag if the packages are all installed
      directly in the prefix or if each package is installed in a subdirectory
      named after the package
    :returns: A mapping from the package name to the set of runtime
      dependencies
    :rtype: dict
    """
    packages = {}
    # since importing colcon_core isn't feasible here the following constant
    # must match colcon_core.location.get_relative_package_index_path()
    subdirectory = 'share/colcon-core/packages'
    if merged_install:
        # return if workspace is empty
        if not (prefix_path / subdirectory).is_dir():
            return packages
        # find all files in the subdirectory
        for p in (prefix_path / subdirectory).iterdir():
            if not p.is_file():
                continue
            if p.name.startswith('.'):
                continue
            add_package_runtime_dependencies(p, packages)
    else:
        # for each subdirectory look for the package specific file
        for p in prefix_path.iterdir():
            if not p.is_dir():
                continue
            if p.name.startswith('.'):
                continue
            p = p / subdirectory / p.name
            if p.is_file():
                add_package_runtime_dependencies(p, packages)

    # remove unknown dependencies
    pkg_names = set(packages.keys())
    for k in packages.keys():
        packages[k] = {d for d in packages[k] if d in pkg_names}

    return packages


def add_package_runtime_dependencies(path, packages):
    """
    Check the path and if it exists extract the packages runtime dependencies.

    :param Path path: The resource file containing the runtime dependencies
    :param dict packages: A mapping from package names to the sets of runtime
      dependencies to add to
    """
    content = path.read_text()
    dependencies = set(content.split(os.pathsep) if content else [])
    packages[path.name] = dependencies


def order_packages(packages):
    """
    Order packages topologically.

    :param dict packages: A mapping from package name to the set of runtime
      dependencies
    :returns: The package names
    :rtype: list
    """
    # select packages with no dependencies in alphabetical order
    to_be_ordered = list(packages.keys())
    ordered = []
    while to_be_ordered:
        pkg_names_without_deps = [
            name for name in to_be_ordered if not packages[name]]
        if not pkg_names_without_deps:
            reduce_cycle_set(packages)
            raise RuntimeError(
                'Circular dependency between: ' + ', '.join(sorted(packages)))
        pkg_names_without_deps.sort()
        pkg_name = pkg_names_without_deps[0]
        to_be_ordered.remove(pkg_name)
        ordered.append(pkg_name)
        # remove item from dependency lists
        for k in list(packages.keys()):
            if pkg_name in packages[k]:
                packages[k].remove(pkg_name)
    return ordered


def reduce_cycle_set(packages):
    """
    Reduce the set of packages to the ones part of the circular dependency.

    :param dict packages: A mapping from package name to the set of runtime
      dependencies which is modified in place
    """
    last_depended = None
    while len(packages) > 0:
        # get all remaining dependencies
        depended = set()
        for pkg_name, dependencies in packages.items():
            depended = depended.union(dependencies)
        # remove all packages which are not dependent on
        for name in list(packages.keys()):
            if name not in depended:
                del packages[name]
        if last_depended:
            # if remaining packages haven't changed return them
            if last_depended == depended:
                return packages.keys()
        # otherwise reduce again
        last_depended = depended


if __name__ == '__main__':  # pragma: no cover
    main()



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import warnings

from colcon_core.package_identification import logger
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version


class PythonPackageIdentification(PackageIdentificationExtensionPoint):
    """
    Identify Python packages with `setup.cfg` files.

    Only packages which pass no arguments (or only a ``cmdclass``) to the
    ``setup()`` function in their ``setup.py`` file are being considered.
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def identify(self, desc):  # noqa: D102
        if desc.type is not None and desc.type != 'python':
            return

        setup_py = desc.path / 'setup.py'
        if not setup_py.is_file():
            return

        setup_cfg = desc.path / 'setup.cfg'
        if not setup_cfg.is_file():
            return

        if not is_reading_cfg_sufficient(setup_py):
            logger.debug(
                "Python package in '{desc.path}' passes arguments to the "
                'setup() function which requires a different identification '
                "extension than '{self.PACKAGE_IDENTIFICATION_NAME}'"
                .format_map(locals()))
            return

        config = get_configuration(setup_cfg)
        name = config.get('metadata', {}).get('name')
        if not name:
            return

        desc.type = 'python'
        if desc.name is not None and desc.name != name:
            msg = 'Package name already set to different value'
            logger.error(msg)
            raise RuntimeError(msg)
        desc.name = name


def is_reading_cfg_sufficient(setup_py):
    """
    Check the content of the setup.py file.

    If the ``setup()`` function is called with no arguments or only a
    ``cmdclass`` it is sufficient to only read the content of the ``setup.cfg``
    file.

    :param setup_py: The path of the setup.py file
    :returns: The flag if reading the setup.cfg file is sufficient
    :rtype: bool
    """
    setup_py_content = setup_py.read_text()
    # the setup function must be called with no arguments
    # or only a ``cmdclass``to be considered by this extension otherwise
    # only reading the content of the setup.cfg file isn't sufficient
    return 'setup()' in setup_py_content or \
        'setup(cmdclass=cmdclass)' in setup_py_content


def get_configuration(setup_cfg):
    """
    Read the setup.cfg file.

    :param setup_cfg: The path of the setup.cfg file
    :returns: The configuration data
    :rtype: dict
    """
    try:
        # import locally to allow other functions in this module to be usable
        from setuptools.config import read_configuration
    except ImportError as e:
        from pkg_resources import get_distribution
        from pkg_resources import parse_version
        setuptools_version = get_distribution('setuptools').version
        minimum_version = '30.3.0'
        if parse_version(setuptools_version) < parse_version(minimum_version):
            e.msg += ', ' \
                "'setuptools' needs to be at least version " \
                '{minimum_version}, if a newer version is not available ' \
                "from the package manager use 'pip3 install -U setuptools' " \
                'to update to the latest version'.format_map(locals())
        raise
    return read_configuration(str(setup_cfg))


def extract_dependencies(options):
    """
    Get the dependencies of the package.

    This function has been depreated, use
    ``colcon_core.package_augmentation.python.extract_dependencies()``
    instead.

    :param options: The dictionary from the options section of the setup.cfg
      file
    :returns: The dependencies
    :rtype: dict(string, set(DependencyDescriptor))
    """
    warnings.warn(
        "'colcon_core.package_identification.python.extract_dependencies()' "
        'has been deprecated, use '
        "'colcon_core.package_augmentation.python.extract_dependencies()' "
        'instead', stacklevel=2)
    from colcon_core.package_augmentation.python import \
        extract_dependencies as function
    return function(options)


def create_dependency_descriptor(requirement_string):
    """
    Create a DependencyDescriptor from a PEP440 compliant string.

    See https://www.python.org/dev/peps/pep-0440/#version-specifiers

    This function has been depreated, use
    ``colcon_core.package_augmentation.python.create_dependency_descriptor()``
    instead.

    :param str requirement_string: a PEP440 compliant requirement string
    :return: A descriptor with version constraints from the requirement string
    :rtype: DependencyDescriptor
    """
    warnings.warn(
        "'colcon_core.package_identification.python."
        "create_dependency_descriptor()' has been deprecated, use "
        "'colcon_core.package_augmentation.python."
        "create_dependency_descriptor()' instead", stacklevel=2)
    from colcon_core.package_augmentation.python import \
        create_dependency_descriptor as function
    return function(requirement_string)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.plugin_system import get_first_line_doc
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name
from colcon_core.plugin_system import order_extensions_by_priority
from colcon_core.plugin_system import order_extensions_grouped_by_priority
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


def test_instantiate_extensions():
    class Extension1:
        pass

    class Extension2:
        pass

    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        # successful instantiation of extensions
        extensions = instantiate_extensions('group')
        assert 'extension1' in extensions.keys()
        assert isinstance(extensions['extension1'], Extension1)
        assert 'extension2' in extensions.keys()
        assert isinstance(extensions['extension2'], Extension2)

        # unique extension instances
        unique_extensions = instantiate_extensions(
            'group', unique_instance=True)
        assert 'extension1' in unique_extensions.keys()
        assert isinstance(unique_extensions['extension1'], Extension1)
        assert extensions['extension1'] != unique_extensions['extension1']

        # exclude extension names
        extensions = instantiate_extensions(
            'group', exclude_names=['extension1'])
        assert 'extension1' not in extensions.keys()
        assert 'extension2' in extensions.keys()


def test_instantiate_extensions_exception():
    class ExtensionRaisingException:

        def __init__(self):
            raise Exception('extension raising exception')

    class ExtensionSkipExtensionException:

        def __init__(self):
            raise SkipExtensionException(
                'extension raising skip extension exception')

    with EntryPointContext(
        exception=ExtensionRaisingException,
        skip_extension_exception=ExtensionSkipExtensionException
    ):
        with patch('colcon_core.plugin_system.logger.error') as error:
            with patch('colcon_core.plugin_system.logger.info') as info:
                extensions = instantiate_extensions('group')

                # the entry point raising an exception different than a skip
                # extension exception results in an error message in the log
                assert error.call_count == 1
                assert len(error.call_args[0]) == 1
                assert "Exception instantiating extension 'group.exception'" \
                    in error.call_args[0][0]
                assert 'extension raising exception' in error.call_args[0][0]

                # the entry point raising a skip extension exception results in
                # an info message in the log
                assert info.call_count == 1
                assert len(info.call_args[0]) == 1
                assert "Skipping extension 'group.skip_extension_exception'" \
                    in info.call_args[0][0]
                assert 'extension raising skip extension exception' \
                    in info.call_args[0][0]
        # neither of the entry points was loaded successfully
        assert extensions == {}


class ExtensionA:
    PRIORITY = 100


class ExtensionB:
    PRIORITY = 100


class ExtensionC:
    PRIORITY = 110


def test_order_extensions_by_name():
    with EntryPointContext(foo=ExtensionA, bar=ExtensionB, baz=ExtensionC):
        extensions = instantiate_extensions('group')
    # ensure correct order based on name
    ordered_extensions = order_extensions_by_name(extensions)
    assert list(ordered_extensions.keys()) == ['bar', 'baz', 'foo']


def test_order_extensions_by_priority():
    with EntryPointContext(foo=ExtensionA, bar=ExtensionB, baz=ExtensionC):
        extensions = instantiate_extensions('group')
    # ensure correct order based on priority
    ordered_extensions = order_extensions_by_priority(extensions)
    assert list(ordered_extensions.keys()) == ['baz', 'bar', 'foo']


def test_order_extensions_grouped_by_priority():
    with EntryPointContext(foo=ExtensionA, bar=ExtensionB, baz=ExtensionC):
        extensions = instantiate_extensions('group')
    # ensure correct order based on priority
    grouped_extensions = order_extensions_grouped_by_priority(extensions)
    assert list(grouped_extensions.keys()) == [110, 100]
    # ensure correct order in each priority group based on name
    assert list(grouped_extensions[110].keys()) == ['baz']
    assert list(grouped_extensions[100].keys()) == ['bar', 'foo']


def test_get_first_line_doc():
    def single_line_doc():
        """Single line."""
    assert get_first_line_doc(single_line_doc) == 'Single line'

    def multi_line_doc():  # noqa: D400
        """
        First line

        Second line.
        """
    assert get_first_line_doc(multi_line_doc) == 'First line'

    def no_doc():
        pass  # pragma: no cover
    assert get_first_line_doc(no_doc) == ''

    def whitespace_doc():
        """ """
    assert get_first_line_doc(whitespace_doc) == ''

    def empty_lines_doc():
        """
        """
    assert get_first_line_doc(empty_lines_doc) == ''


def test_satisfies_version():
    satisfies_version('1.2.3', '^1')
    satisfies_version('1.2.3', '^1.1')

    with pytest.raises(RuntimeError) as e:
        satisfies_version('1.0.3', '^1.1')
    assert 'too old' in str(e.value)

    with pytest.raises(RuntimeError) as e:
        satisfies_version('2.0.0', '^1.2')
    assert 'newer' in str(e.value)

    # different semantic for version numbers before 1.0
    with pytest.raises(RuntimeError) as e:
        satisfies_version('0.2.3', '^0.1')
    assert 'newer' in str(e.value)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

"""
Call a subprocess making the stdout and stderr output available via callbacks.

The stdout and stderr pipes are read concurrently using the asyncio event loop
to maintain the original order as closely as possible.
"""

import asyncio
from concurrent.futures import ALL_COMPLETED
from functools import partial
import os
import platform
import shlex
import subprocess
import sys
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence

from colcon_core.logging import colcon_logger

SIGINT_RESULT = 'SIGINT'

logger = colcon_logger.getChild(__name__)


def new_event_loop():
    """
    Create a new event loop.

    On Windows return a ProactorEventLoop.

    :returns: The created event loop
    """
    if sys.platform == 'win32':
        return asyncio.ProactorEventLoop()
    return asyncio.new_event_loop()


async def run(
    args: Sequence[str],
    stdout_callback: Callable[[bytes], None],
    stderr_callback: Callable[[bytes], None],
    *,
    cwd: str = None,
    env: Mapping[str, str] = None,
    shell: bool = False,
    use_pty: Optional[bool] = None
) -> subprocess.CompletedProcess:
    """
    Run the command described by args.

    Invokes the callbacks for every line read from the subprocess pipes.

    :param args: args should be a sequence of program arguments
    :param stdout_callback: the callable is invoked for every line read from
      the stdout pipe of the process
    :param stderr_callback: the callable is invoked for every line read from
      the stderr pipe of the process
    :param cwd: the working directory for the subprocess
    :param env: a dictionary with environment variables
    :param shell: whether to use the shell as the program to execute
    :param use_pty: whether to use a pseudo terminal
    :returns: the result of the completed process
    :rtype subprocess.CompletedProcess
    """
    assert callable(stdout_callback) or stdout_callback is None
    assert callable(stderr_callback) or stderr_callback is None

    # if use_pty is neither True nor False choose based on isatty of stdout
    if use_pty is None:
        use_pty = sys.stdout.isatty()
    # the pty module is only supported on Windows
    if use_pty and platform.system() != 'Linux':
        use_pty = False

    rc, _, _ = await _async_check_call(
        args, stdout_callback, stderr_callback,
        cwd=cwd, env=env, shell=shell, use_pty=use_pty)
    return subprocess.CompletedProcess(args, rc)


async def check_output(
    args: Sequence[str],
    *,
    cwd: str = None,
    env: Mapping[str, str] = None,
    shell: bool = False
) -> subprocess.CompletedProcess:
    """
    Get the output of an invoked command.

    :param args: args should be a sequence of program arguments
    :param cwd: the working directory for the subprocess
    :param env: a dictionary with environment variables
    :param shell: whether to use the shell as the program to execute
    :returns: The `stdout` output of the command
    :rtype: str
    """
    rc, stdout_data, stderr_data = await _async_check_call(
        args, subprocess.PIPE, subprocess.PIPE,
        cwd=cwd, env=env, shell=shell, use_pty=False)
    if rc:
        stderr_data = stderr_data.decode(errors='replace')
    assert not rc, \
        'Expected {args} to pass: {stderr_data}'.format_map(locals())
    return stdout_data


async def _async_check_call(
    args, stdout_callback, stderr_callback, *,
    cwd=None, env=None, shell=False, use_pty=None
):
    """Coroutine running the command and invoking the callbacks."""
    # choose function to create subprocess
    if not shell:
        create_subprocess = asyncio.create_subprocess_exec
    else:
        args = [' '.join([escape_shell_argument(a) for a in args])]
        create_subprocess = asyncio.create_subprocess_shell

    # choose stdout and stderr arguments for the subprocess
    stdout = subprocess.PIPE if stdout_callback else subprocess.DEVNULL
    stderr = subprocess.PIPE if stderr_callback else subprocess.DEVNULL

    # open pseudo terminals
    if use_pty:
        # only import when requested since it is not available on all platforms
        import pty
        if stdout_callback:
            stdout_descriptor, stdout = pty.openpty()
        if stderr_callback:
            stderr_descriptor, stderr = pty.openpty()

    process = await create_subprocess(
        *args, cwd=cwd, env=env, stdout=stdout, stderr=stderr)

    # read pipes concurrently
    callbacks = []
    if use_pty:
        if callable(stdout_callback):
            callbacks.append(_fd2callback(stdout_descriptor, stdout_callback))
        if callable(stderr_callback):
            callbacks.append(_fd2callback(stderr_descriptor, stderr_callback))
    else:
        if callable(stdout_callback):
            callbacks.append(_pipe2callback(
                process.stdout, stdout_callback,
                process.stderr if callable(stderr_callback) else None))
        if callable(stderr_callback):
            callbacks.append(asyncio.ensure_future(_pipe2callback(
                process.stderr, stderr_callback,
                process.stdout if callable(stdout_callback) else None)))

    output = [None, None]
    if not stdout_callback and not stderr_callback:
        # asynchronously wait for the subprocess
        await process.wait()
    else:
        # asynchronously communicate with the subprocess
        callbacks.append(process.wait())
        if subprocess.PIPE in (stdout_callback, stderr_callback):
            callbacks.append(_communicate_and_close_fds(
                process,
                # collect output in case the process uses any pipes
                output,
                # pseudo terminals need to be closed explicitly
                stdout if use_pty else None, stderr if use_pty else None))
        else:
            callbacks.append(_wait_and_close_fds(
                process,
                # pseudo terminals need to be closed explicitly
                stdout if use_pty else None, stderr if use_pty else None))
        # waiting for coroutines is deprecated as of Python 3.8
        # convert coroutines into tasks
        for i, callback in enumerate(callbacks):
            if not isinstance(callback, asyncio.Task):
                try:
                    callbacks[i] = asyncio.create_task(callback)
                except AttributeError:
                    # fallback for Python < 3.7
                    callbacks[i] = asyncio.ensure_future(callback)
        try:
            done, _ = await asyncio.wait(callbacks, return_when=ALL_COMPLETED)
        except (asyncio.CancelledError, KeyboardInterrupt):
            # finish the communication with the subprocess
            done, _ = await asyncio.wait(callbacks, return_when=ALL_COMPLETED)
            raise
        finally:
            # read potential exceptions to avoid asyncio errors
            for task in done:
                _ = task.exception()  # noqa: F841

    return process.returncode, output[0], output[1]


def escape_shell_argument(arg):
    """
    Escape the shell arguments for an invocation through a shell.

    :param arg: A single command line argument
    :returns: The escaped command line argument
    :rtype: str
    """
    # some literals must not be quoted
    unquoted_values = [';', '|', '&&', '||']
    if arg in unquoted_values:
        return arg
    # some arguments don't need quoting
    if arg.startswith('`') and arg.endswith('`'):
        return arg
    if arg.startswith('$(') and arg.endswith(')'):
        return arg

    quoted = shlex.quote(arg)
    if sys.platform == 'win32':
        # Windows doesn't like paths with single quotes
        if len(quoted) > 1 and quoted.startswith("'") and quoted.endswith("'"):
            quoted = '"' + quoted[1:-1] + '"'

    return quoted


async def _fd2callback(descriptor, callback):
    """Coroutine reading from fd and invoking the callback for each line."""
    func = partial(_blocking_fd2callback, descriptor, callback)
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, func)


def _blocking_fd2callback(descriptor, callback):
    """Read all lines from the stream invoke the callback for each line."""
    with os.fdopen(descriptor) as stream:
        while True:
            try:
                line = stream.readline()
            except IOError:
                # this is how the fd signals the EOF
                break
            callback(line.encode())


async def _pipe2callback(stream, callback, other_stream=None):
    """Coroutine reading from pipe and invoking the callback for each line."""
    while True:
        line = await stream.readline()
        if not line:
            # this is how the pipe signals the EOF
            break
        callback(line)

    # HACK on Windows sometimes only one of the two streams gets closed
    # feeding an EOF explicitly ensures that the other coroutine finishes
    if sys.platform == 'win32' and other_stream:
        other_stream.feed_eof()


async def _wait_and_close_fds(process, stdout=None, stderr=None):
    """Coroutine waiting for the process and closing all handles."""
    try:
        await process.wait()
    finally:
        # always close handles even when a CancelledError is raised
        if stdout:
            os.close(stdout)
        if stderr:
            os.close(stderr)


async def _communicate_and_close_fds(
    process, output, stdout=None, stderr=None
):
    """Coroutine communicating with the process and closing all handles."""
    stdout_data, stderr_data = await process.communicate()
    output[0] = stdout_data
    output[1] = stderr_data
    if stdout:
        os.close(stdout)
    if stderr:
        os.close(stderr)
